/***************************************************************************
                          HttpSession.cpp  -  description
                             -------------------
    begin                : Fri May 4 2002
    copyright            : (C) 2002 by Henrik Baastrup
    email                : hba@libero.it
    home page            : http://digilander.iol.it/hba/Linux/index.html
 ***************************************************************************/

/***************************************************************************
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU LGPL (Lesser general Public License) 
 *   as published by the Free Software Foundation:
 *   http://www.gnu.org/copyleft/lesser.html or in the file copying.txt
 *
 ***************************************************************************/
#include <string.h>
#include <time.h>
#include <signal.h>
#include "HttpSession.h"
#include "MiniHTTP.h"

#ifdef LINUX
void *runSessionThread(void *pObj)
#elif defined(WIN32)
DWORD runSessionThread(void *pObj)
#endif
/*
 * This is the thread function that lunch the goSession method in the Session class
 */
{
	HttpSession *thisSession = (HttpSession *)pObj;
	do {
		thisSession->goSession();
	} while (thisSession->isConnectionKeepAlive());

	thisSession->nullSessionThreadId(); // Do not kill this thread when delete object
	delete thisSession;  // Do suicide
#ifdef LINUX
	pthread_exit(NULL);
#elif defined(WIN32)
	ExitThread(0);
#endif
	return 0;
}//runSessionThread



HttpSession::HttpSession(MiniHTTP *http, struct sockaddr_in *clientAddr, int clientAddrLen)
{
	m_miniHTTP = http;
	m_clientSocketFd = -1;
	m_nextSession = NULL;
	m_threadId = 0;
	strcpy(m_serverName, "MiniHTTP/1.0");
	*m_httpMethod = '\0';
	*m_requestURI = '\0';
	*m_contentType = '\0';
	strcpy(m_httpSupportedVersion,"HTTP/1.0");
	m_contentLength = 0;
	m_connectionKeepAlive = 0;
	m_touchTime = time(NULL);
	if (clientAddr==NULL || clientAddrLen<=0) m_clientAddr = NULL;
	else {
		m_clientAddr = (struct sockaddr_in *)malloc(clientAddrLen);
		memcpy(m_clientAddr, clientAddr, clientAddrLen);
	}
#ifdef LINUX
	pthread_attr_init(&m_threadAttr);
#endif
}//HttpSession


HttpSession::~HttpSession()
{
#ifdef LINUX
	if (m_clientSocketFd >= 0) close(m_clientSocketFd);
	if (m_threadId > 0) pthread_kill(m_threadId, SIGKILL); // Kill session thread
	pthread_attr_destroy(&m_threadAttr);
#elif defined(WIN32)
	if (m_clientSocketFd >= 0) closesocket(m_clientSocketFd);
	if (m_threadId) CloseHandle(m_threadId); // Kill session thread
#endif
	shutdown(m_clientSocketFd, 2);
	m_clientSocketFd = -1;
	m_threadId = 0;
	if (m_miniHTTP) m_miniHTTP->deleteSession(this);
	if (m_clientAddr) free(m_clientAddr);
}//~HttpSession


void HttpSession::runSession(SOCKET sockFd)
/*
 * This method is callded to lunch a new thread for the Session class.
 */
{
	m_clientSocketFd = sockFd;
#ifdef LINUX
	int r = pthread_attr_setdetachstate(&m_threadAttr, PTHREAD_CREATE_DETACHED);
	r = pthread_create(&m_threadId, &m_threadAttr, runSessionThread, (void *) this);
#elif defined(WIN32)
	DWORD id;
	m_threadId = CreateThread(0, 0, (LPTHREAD_START_ROUTINE) runSessionThread, (void *)this, 0, &id);
#endif
}//runSession


void HttpSession::goSession()
/*
 * This method is the actual thread method
 */
{
	fd_set readFdSet;
//	struct timeval tv;
	char logStr[1024], buffer[1024], *ptr, *bodyPtr, *respPtr, *cp;
	unsigned char *host;
	int nLeft, nRead, nBodyRead, gotHeader, rVal, i;

	if (m_clientAddr==NULL) m_miniHTTP->logWrite(1,"ERROR Contacted by an unkonwn host");
	else {
		host = (unsigned char *)&m_clientAddr->sin_addr;
		sprintf(logStr,"Contacted by host: %d.%d.%d.%d",host[0],host[1],host[2],host[3]);
		m_miniHTTP->logWrite(1,logStr);
	}
	if (m_clientSocketFd < 0) {
		m_connectionKeepAlive = 0;
		return;
	}

  // Read the HTTP header
	FD_ZERO(&readFdSet);
	buffer[1023] = '\0'; //stop
	nLeft = 1023;
	nBodyRead = -1;
	ptr = buffer;
	gotHeader = 0;
	while (nLeft > 0) {
//		tv.tv_sec = 1;
//		tv.tv_usec = 0;
//		FD_SET(m_clientSocketFd, &readFdSet);
//		if ((nRead=select(m_clientSocketFd+1, &readFdSet, NULL, NULL, &tv))==0) break;
		nRead  = recv(m_clientSocketFd, ptr, nLeft, 0);
//		nRead  = read(m_clientSocketFd, ptr, nLeft);
		if (nRead < 0) {
			m_connectionKeepAlive = 0;
			break;
		}
		else if (nRead==0) {
			m_connectionKeepAlive = 0;
			break;
		}
		cp = ptr;
		if (gotHeader>=0) for (i=0; i<nRead; i++) {
			if (*cp=='\r') gotHeader++;
			else if (*cp=='\n') gotHeader++;
			else gotHeader = 0;
			if (gotHeader >= 4) break;
			cp++;
		}
		nLeft -= nRead;
		if (nBodyRead >= 0) nBodyRead += nRead;
		ptr += nRead;
		if (gotHeader>=4) { // We have a HTTP header, parse it!
			gotHeader = -1;
			*ptr = '\0';
			sprintf(logStr,"HEADER from host: %d.%d.%d.%d\n%s",host[0],host[1],host[2],host[3],buffer);
			m_miniHTTP->logWrite(3,logStr);
			if ((bodyPtr=parseHttpHeader(buffer))==NULL) break;
			nBodyRead = strlen(bodyPtr);
		}
		if (m_contentLength>0 && (nBodyRead)>=m_contentLength) break;
		if (m_contentLength==0 && gotHeader<0) break;
	}
	*ptr = '\0';

	// We have a HTTP header and body
	m_contentLength = 0;
    if (bodyPtr==NULL) {
		sprintf(logStr,"ERROR 400 Bad Request from host: %d.%d.%d.%d",host[0],host[1],host[2],host[3]);
		m_miniHTTP->logWrite(2,logStr);
		sprintf(buffer,"%s 400 Bad Request\r\nServer: %s\r\nContent-type: text/html\r\n\r\n",m_httpSupportedVersion,m_serverName);
		sendStr(buffer);
		sprintf(buffer,"<HTML><HEAD</HEAD<BODY>\n<H1>Error</H1>/n400 Bad Request\n</BODY></HTML>\n");
		sendStr(buffer);
	}
	else if (strcmp(m_httpMethod, "GET")==0) { // Is the called HTTP method GET?
		sprintf(logStr,"GET uri=%s ver=%s from host: %d.%d.%d.%d",m_requestURI,m_httpRequestVersion,host[0],host[1],host[2],host[3]);
		m_miniHTTP->logWrite(2,logStr);
		setContentType("text/html");
		if (strcmp(m_requestURI,"/do?shutdown=1")==0) {// Shutdown the http service!
			if (m_miniHTTP!=NULL) {
				m_miniHTTP->stopMiniHttp();
				sprintf(buffer,"%s 200 OK\r\nServer: %s\r\nContent-type: text/html\r\n\r\n",m_httpSupportedVersion,m_serverName);
				sendStr(buffer);
				sprintf(buffer,"<HTML><HEAD</HEAD<BODY>\n<H1>Shutdown</H1>\nThe Server goes down now!\n</BODY></HTML>\n");
				sendStr(buffer);
			}
		}
		else callGetFunction(bodyPtr);
	}
	else if (strcmp(m_httpMethod, "POST")==0) { // Is the called HTTP method POST?
		sprintf(logStr,"POST uri=%s from host: %d.%d.%d.%d",m_requestURI,host[0],host[1],host[2],host[3]);
		m_miniHTTP->logWrite(2,logStr);
		if (m_contentLength>0) ptr[m_contentLength] = '\0';
		setContentType("text/html");
		// Do POST!
		respPtr = NULL;
		rVal = callPostFunction(bodyPtr, &respPtr);
		if (rVal != 0) {
			// An error occured. Respond with error!
			if (respPtr!=NULL) {
				sprintf(buffer,"%s %s\r\nServer: %s\r\nContent-type: text/html\r\n\r\n",m_httpSupportedVersion,respPtr,m_serverName);
				sendStr(buffer);
				sprintf(buffer,"<HTML><HEAD</HEAD<BODY>\n<H1>Error</H1>\n%s\n</BODY></HTML>\n",respPtr);
				sendStr(buffer);
				free(respPtr);
			}
			else  {
				sprintf(buffer,"%s 500\r\nServer: %s\r\nContent-type: text/html\r\n\r\n",m_httpSupportedVersion,m_serverName);
				sendStr(buffer);
				sprintf(buffer,"<HTML><HEAD</HEAD<BODY>\n<H1>Error</H1>\nInternal Server Error\n</BODY></HTML>\n");
				sendStr(buffer);
			}
		}
		else {
			if (respPtr != NULL) {
				// Respond with the requested POST data
				m_contentLength = strlen(respPtr);
				writeOkHeader();
				sendStr(respPtr);
				free(respPtr);
			}
			else {
				sprintf(buffer,"%s 200\r\nServer: %s\r\nContent-type: text/html\r\n\r\n\r\n",m_httpSupportedVersion,m_serverName);
				sendStr(buffer);
			}
		}
	}
	else { //Not supported HTTP method!
		sprintf(logStr,"ERROR 405 Method %s Not Allowed from host: %d.%d.%d.%d",m_httpMethod,host[0],host[1],host[2],host[3]);
		m_miniHTTP->logWrite(2,logStr);
		sprintf(buffer,"%s 405 Method Not Allowed\r\nAllow: GET, PUT\r\nServer: %s\r\nContent-type: text/html\r\n\r\n",m_httpSupportedVersion,m_serverName);
		sendStr(buffer);
		sprintf(buffer,"<HTML><HEAD</HEAD<BODY>\n<H1>Error</H1>\n405 Method Not Allowedt\n</BODY></HTML>\n");
		sendStr(buffer);
	}

	if (m_connectionKeepAlive) m_connectionKeepAlive = 0; // I have not implemented the Keep-Alive function jet!
#ifdef LINUX
	close(m_clientSocketFd);
#elif defined(WIN32)
	closesocket(m_clientSocketFd);
#endif
//	shutdown(m_clientSocketFd, 2);
	m_clientSocketFd = -1;
}//goSession


void HttpSession::writeOkHeader()
{
	char buffer[1024];
	char timeStr[80];
	struct tm *tim;
	time_t now;

	time(&now);
	tim = localtime(&now);
	strcpy(timeStr,asctime(tim)); timeStr[strlen(timeStr)-1] = '\0';

	sprintf(buffer,"%s 200 OK\r\nServer: %s\r\nLast-Modified: %s\r\nContent-Type: %s\r\n",m_httpSupportedVersion,m_serverName,timeStr,m_contentType);
	sendStr(buffer);
	if (m_contentLength > 0) {
		sprintf(buffer,"Content-Length: %ld\r\n",m_contentLength);
		sendStr(buffer);
	}
	sprintf(buffer,"\r\n\n");
	sendStr(buffer);
}//writeOkHeader


void HttpSession::writeError(char *headerStr, char *bodyStr)
{
	char buffer[1024];

	sprintf(buffer,"%s %s\r\nServer: %s\r\nContent-type: %s\r\n\r\n",m_httpSupportedVersion,headerStr,m_serverName,m_contentType);
	sendStr(buffer);
	sendStr(bodyStr);
}//writeError



char *HttpSession::parseHttpHeader(char *buffer)
/*
 * This metod is a parser for the HTTP header
 */
{
	int phase, subPhase, stop;
	char *ptr, *srcPtr, *upPtr;

	if (buffer==NULL) return NULL;

	stop = 0;
	phase = 1;
	subPhase = 1;
	srcPtr = ptr = buffer;
	while (*ptr && ! stop) {
		switch (*ptr) {
			case '\r':
				if (*(ptr+1)!='\n') return NULL;
				*ptr = '\0';
				ptr += 2;
				if (*ptr=='\r') { //empty line seprator between header and body
					while (*ptr=='\r' || *ptr=='\n') ptr++;
					stop = 1;
				}
				switch (subPhase) {
					case 1:
						strncpy(m_httpRequestVersion, srcPtr, 15);
						m_httpRequestVersion[14] = '\0';
						if (strncmp(m_httpRequestVersion, "HTTP/1.1", 8)==0) m_connectionKeepAlive = 1;
						else m_connectionKeepAlive = 0;
						break;
					case 2:
						strncpy(m_contentType, srcPtr, 256);
						m_contentType[255] = '\0';
						break;

					case 3:
						m_contentLength = atol(srcPtr);
						break;
					case 4:
						if (strncmp(srcPtr, "Keep-Alive", 10)==0) m_connectionKeepAlive = 1;
						else if (strncmp(srcPtr, "Close", 5)==0) m_connectionKeepAlive = 0;
						break;
				}
				if (stop) return ptr;
				srcPtr = ptr;
				break;

			case ' ':
				*ptr = '\0';
				if (phase==3) {
					upPtr = srcPtr;
					while(*upPtr) {
						if (*upPtr >='A' && *upPtr <= 'Z') *upPtr += 'a' - 'A';
						upPtr++;
					}
				}
				switch (phase) {
					case 1: //Read the method (GET, POST)
						strncpy(m_httpMethod, srcPtr, 10);
						m_httpMethod[9] = '\0';
						srcPtr = ptr;
						srcPtr++;
						phase = 2;
						break;

					case 2: //Read the URI
						strncpy(m_requestURI, srcPtr, 1024);
						m_requestURI[1023] = '\0';
						srcPtr = ptr;
						srcPtr++;
						phase = 3;
						break;

					case 3: //Read the rest of the header
	m_miniHTTP->logWrite(1,"Some more header info");
						if (strcmp(srcPtr, "content-type:")==0) subPhase = 2;
						else if (strcmp(srcPtr, "content-length:")==0) subPhase = 3;
						else if (strcmp(srcPtr, "connection:")==0) subPhase = 4;
						else subPhase = 100;
						srcPtr = ptr;
						srcPtr++;
				}
				break;
		}
		ptr++;
	}

	return ptr;
}//parseHttp


int HttpSession::sendStr(char *str)
/*
 * This method send a string to the client socket
 */
{
	int nleft, nwritten;
	char *ptr;

	if (str==NULL) return 0;

//	nleft = strlen(str)+1;
	nleft = strlen(str);
	ptr = str;
	while (nleft > 0) {
//		nwritten = write(m_clientSocketFd, ptr, nleft);
		nwritten = send(m_clientSocketFd, ptr, nleft, 0);
		if (nwritten < 0) return -1;
		nleft -= nwritten;
		ptr += nwritten;
	}
	return nleft;
}//sendStr


//inline char *Session::getHomeDir()
char *HttpSession::getHomeDir()
{
	if (m_miniHTTP) return m_miniHTTP->getHomeDir();
	else return NULL;
}//getHomeDir


void HttpSession::callGetFunction(char *msg)
{
	if (m_miniHTTP) m_miniHTTP->callGetFunction(msg, this);
}//callGetFunction


int HttpSession::callPostFunction(char *msg, char **rsp)
{
	if (m_miniHTTP) return m_miniHTTP->callPostFunction(msg, rsp, this);
	return 0;
}//callPostFunction
